#!/usr/bin/python
# service_supplier.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_supplier.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_supplier
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return SupplierService()


class SupplierService(LocalService):

    service_name = 'service_supplier'
    config_path = 'services/supplier/enabled'

    def dependent_on(self):
        return [
            'service_keys_registry',
        ]

    def installed(self):
        from bitdust.userid import my_id
        if not my_id.isLocalIdentityReady():
            return False
        return True

    def attached_dht_layers(self):
        from bitdust.dht import dht_records
        return [
            dht_records.LAYER_SUPPLIERS,
        ]

    def start(self):
        from bitdust.supplier import customer_space
        customer_space.init()
        return True

    def stop(self):
        from bitdust.supplier import customer_space
        customer_space.shutdown()
        return True

    def request(self, json_payload, newpacket, info):
        from bitdust.supplier import customer_space
        return customer_space.on_service_supplier_request(json_payload, newpacket, info)

    def cancel(self, json_payload, newpacket, info):
        from bitdust.supplier import customer_space
        return customer_space.on_service_supplier_cancel(json_payload, newpacket, info)



#!/usr/bin/python
# service_blockchain_explorer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_blockchain_explorer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_blockchain_explorer
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return BlockchainExplorerService()


class BlockchainExplorerService(LocalService):

    service_name = 'service_blockchain_explorer'
    config_path = 'services/blockchain-explorer/enabled'

    def dependent_on(self):
        return [
            'service_bismuth_node',
        ]

    def installed(self):
        return True

    def start(self):
        from bitdust.blockchain import blockchain_explorer
        return blockchain_explorer.init()

    def stop(self):
        from bitdust.blockchain import blockchain_explorer
        return blockchain_explorer.shutdown()



#!/usr/bin/python
# service_supplier_contracts.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_supplier_contracts.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_supplier_contracts
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return SupplierContractsService()


class SupplierContractsService(LocalService):

    service_name = 'service_supplier_contracts'
    config_path = 'services/supplier-contracts/enabled'

    def dependent_on(self):
        return [
            'service_supplier',
            'service_blockchain_id',
        ]

    def start(self):
        from bitdust.main import events
        from bitdust.supplier import storage_contract
        events.add_subscriber(self.on_blockchain_transaction_received, 'blockchain-transaction-received')
        storage_contract.scan_recent_storage_transactions()
        return True

    def stop(self):
        from bitdust.main import events
        events.remove_subscriber(self.on_blockchain_transaction_received, 'blockchain-transaction-received')
        return True

    def on_blockchain_transaction_received(self, evt):
        if evt.data.get('operation') == 'storage':
            from bitdust.supplier import storage_contract
            storage_contract.verify_accept_storage_payment(evt.data)



#!/usr/bin/python
# web_socket_transmitter.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (web_socket_transmitter.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: web_socket_transmitter

Secure remote routed connection to access BitDust nodes from mobile devices.
"""

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os

#------------------------------------------------------------------------------

from twisted.application.strports import listen
from twisted.internet.protocol import Protocol, Factory

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import txws
from bitdust.lib import strng
from bitdust.lib import serialization
from bitdust.lib import jsn
from bitdust.lib import utime

from bitdust.system import local_fs

from bitdust.crypt import cipher

from bitdust.main import config
from bitdust.main import settings

#------------------------------------------------------------------------------

MAX_KNOWN_ROUTES = 1000
MAX_ACTIVE_ROUTES = 100

#------------------------------------------------------------------------------

_WebSocketListener = None
_WebSocketTransports = {}
_WebSocketRouterLocation = None
_Routes = {}

#------------------------------------------------------------------------------


def init():
    global _WebSocketListener
    global _WebSocketRouterLocation
    if _WebSocketListener is not None:
        lg.warn('_WebSocketListener already initialized')
        return True
    host = config.conf().getString('services/web-socket-router/host').strip()
    port = config.conf().getInt('services/web-socket-router/port')
    _WebSocketRouterLocation = 'ws://{}:{}'.format(host, port)
    service_dir = settings.ServiceDir('service_web_socket_router')
    if not os.path.exists(service_dir):
        if _Debug:
            lg.out(_DebugLevel, 'web_socket_transmitter.init will create folder: ' + service_dir)
        os.makedirs(service_dir)
    load_routes()
    try:
        ws = WrappedWebSocketFactory(WebSocketFactory())
        _WebSocketListener = listen('tcp:%d' % port, ws)
    except:
        lg.exc()
        return False
    if _Debug:
        lg.out(_DebugLevel, 'web_socket_transmitter.init  _WebSocketListener=%r' % _WebSocketListener)
    return True


def shutdown():
    global _WebSocketListener
    global _WebSocketRouterLocation
    if _WebSocketListener:
        if _Debug:
            lg.out(_DebugLevel, 'web_socket_transmitter.shutdown calling _WebSocketListener.stopListening()')
        _WebSocketListener.stopListening()
        del _WebSocketListener
        _WebSocketListener = None
        if _Debug:
            lg.out(_DebugLevel, '    _WebSocketListener destroyed')
    else:
        lg.warn('_WebSocketListener is None')
    _WebSocketRouterLocation = None
    return True


#------------------------------------------------------------------------------


def location():
    global _WebSocketRouterLocation
    return _WebSocketRouterLocation


def routes(route_id=None):
    global _Routes
    if route_id is None:
        return _Routes
    return _Routes.get(route_id)


#------------------------------------------------------------------------------


def validate_route_id(route_id):
    # SECURITY
    # TODO: add more strict validation of the route_id
    rid = strng.to_text(route_id)
    if len(rid) != 8:
        raise Exception('invalid route_id length')
    return rid


#------------------------------------------------------------------------------


def add_route():
    global _Routes
    route_id = None
    while not route_id or route_id in _Routes:
        route_id = cipher.generate_secret_text(5)
    # this is where Mobile device will be connecting
    route_url = 'ws://{}:{}/?r={}'.format(
        config.conf().getString('services/web-socket-router/host').strip(),
        config.conf().getInt('services/web-socket-router/port'),
        route_id,
    )
    # this is where BitDust node will be connecting
    internal_url = 'ws://{}:{}/?i={}'.format(
        config.conf().getString('services/web-socket-router/host').strip(),
        config.conf().getInt('services/web-socket-router/port'),
        route_id,
    )
    _Routes[route_id] = {
        'route_id': route_id,
        'route_url': route_url,
        'external_transport': None,
        'internal_url': internal_url,
        'internal_transport': None,
        'created': utime.get_sec1970(),
    }
    if _Debug:
        lg.args(_DebugLevel, route_id=route_id, routes=len(_Routes))
    return route_id


def remove_route(route_id):
    global _Routes
    _Routes.pop(route_id, None)
    if _Debug:
        lg.args(_DebugLevel, route_id=route_id, routes=len(_Routes))


#------------------------------------------------------------------------------


def load_routes():
    global _Routes
    service_dir = settings.ServiceDir('service_web_socket_router')
    routes_file_path = os.path.join(service_dir, 'routes')
    _Routes = jsn.loads_text(local_fs.ReadTextFile(routes_file_path)) or {}
    if _Debug:
        lg.args(_DebugLevel, count=len(_Routes))


def save_routes():
    global _Routes
    service_dir = settings.ServiceDir('service_web_socket_router')
    routes_file_path = os.path.join(service_dir, 'routes')
    ret = {}
    for route_id, route_info in _Routes.items():
        ret[route_id] = {
            'route_id': route_info['route_id'],
            'route_url': route_info['route_url'],
            'internal_url': route_info['internal_url'],
            'created': route_info['created'],
        }
    if _Debug:
        lg.args(_DebugLevel, count=len(ret))
    return local_fs.WriteTextFile(routes_file_path, jsn.dumps(ret, indent=2, separators=(',', ':')))


#------------------------------------------------------------------------------


class WrappedWebSocketProtocol(txws.WebSocketProtocol):

    route_id = None
    direction = None

    def connectionMade(self):
        global MAX_ACTIVE_ROUTES
        global _WebSocketTransports
        if len(_WebSocketTransports) >= MAX_ACTIVE_ROUTES:
            lg.warn('maximum number of active routes reached')
            self.loseConnection()
            return
        return txws.WebSocketProtocol.connectionMade(self)

    def validateHeaders(self):
        _, _, parameter = self.location.partition('?')
        if not parameter:
            return txws.WebSocketProtocol.validateHeaders(self)
        param_name, _, route_id = parameter.partition('=')
        try:
            validate_route_id(route_id)
        except:
            lg.exc()
            return False
        if param_name in ('r', 'route', 'i', 'internal'):
            if not routes(route_id):
                lg.warn('rejected connection %r, route %r is unknown' % (self, route_id))
                return False
        else:
            lg.warn('rejected connection %r, invalid input parameters' % self)
            return False
        if route_id:
            self.direction = 'internal' if param_name in ('i', 'internal') else 'external'
            self.route_id = route_id
            if self.route_id in routes():
                routes(self.route_id)[self.direction + '_transport'] = self
                if _Debug:
                    lg.dbg(_DebugLevel, 'registered %s transport %s for route %s' % (self.direction, self, self.route_id))
            else:
                lg.warn('route %s was not registered' % self.route_id)
                return False
        if _Debug:
            lg.args(_DebugLevel, route_id=self.route_id, direction=self.direction, transport=self)
        return txws.WebSocketProtocol.validateHeaders(self)


class WrappedWebSocketFactory(txws.WebSocketFactory):

    protocol = WrappedWebSocketProtocol


#------------------------------------------------------------------------------


class WebSocketProtocol(Protocol):

    _key = None

    def dataReceived(self, raw_data):
        if _Debug:
            lg.args(_DebugLevel, raw_data=raw_data)
        try:
            json_data = serialization.BytesToDict(raw_data, keys_to_text=True, values_to_text=True, encoding='utf-8')
        except:
            lg.exc()
            return
        if _Debug:
            lg.dbg(_DebugLevel, 'received %d bytes from web socket at %r' % (len(raw_data), self._key))
        if not do_process_incoming_message(self.transport, json_data, raw_data):
            lg.warn('failed processing incoming message from web socket: %r' % json_data)
            self.transport.loseConnection()

    def connectionMade(self):
        global _WebSocketTransports
        ret = Protocol.connectionMade(self)
        peer = self.transport.getPeer()
        self._key = (peer.type, peer.host, peer.port)
        peer_text = '%s://%s:%s' % (self._key[0], self._key[1], self._key[2])
        _WebSocketTransports[self._key] = self.transport
        if _Debug:
            lg.args(_DebugLevel, peer=peer_text, transport=self.transport, ws_connections=len(_WebSocketTransports))
        return ret

    def connectionLost(self, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, args=args, kwargs=kwargs)
        global _WebSocketTransports
        Protocol.connectionLost(self, *args, **kwargs)
        _WebSocketTransports.pop(self._key)
        peer_text = '%s://%s:%s' % (self._key[0], self._key[1], self._key[2])
        self._key = None
        route_id = None
        direction = None
        try:
            route_id = self.transport.route_id
            direction = self.transport.direction
        except:
            lg.exc()
        cleaned = False
        if route_id and direction:
            if routes(route_id):
                if routes(route_id).get(direction + '_transport'):
                    routes(route_id)[direction + '_transport'] = None
                    cleaned = True
        if _Debug:
            lg.args(_DebugLevel, peer=peer_text, transport=self.transport, ws_connections=len(_WebSocketTransports), route_id=route_id, direction=direction, cleaned=cleaned)


#------------------------------------------------------------------------------


class WebSocketFactory(Factory):

    protocol = WebSocketProtocol

    def buildProtocol(self, addr):
        proto = Factory.buildProtocol(self, addr)
        return proto


#------------------------------------------------------------------------------


def do_process_incoming_message(transport, json_data, raw_data):
    global MAX_KNOWN_ROUTES
    route_id = transport.route_id
    transport_direction = transport.direction
    if _Debug:
        lg.args(_DebugLevel, route_id=route_id, direction=transport_direction, json_data=json_data)
    if route_id is None and transport_direction is None:
        if json_data.get('cmd') == 'connect-request':
            if len(routes()) >= MAX_KNOWN_ROUTES:
                if _Debug:
                    lg.dbg(_DebugLevel, 'request for a new route from %r was rejected: too many routes are currently registered' % transport)
                return False
            route_id = add_route()
            save_routes()
            route_info = routes(route_id)
            json_response = {
                'result': 'accepted',
                'route_id': route_id,
            }
            response_raw_data = serialization.DictToBytes(json_response, encoding='utf-8')
            try:
                transport.write(response_raw_data)
            except:
                lg.exc()
                return False
            if _Debug:
                lg.out(_DebugLevel, '    wrote %d bytes to %s/%s at %r' % (len(response_raw_data), route_id, transport_direction, transport))
            return True
        return False
    if not route_id:
        lg.warn('unknown route ID: %r' % transport)
        return False
    if not transport_direction:
        lg.warn('route direction was not identified: %r' % transport)
        return False
    cmd = json_data.get('cmd')
    if cmd == 'handshake':
        if json_data.get('internal') and transport_direction == 'internal':
            route_info = routes(route_id)
            route_url = route_info['route_url'] if route_info else None
            if not route_url:
                lg.warn('route info for %r was not found' % route_id)
                return False
            json_response = {
                'cmd': 'handshake-accepted',
                'route_url': route_url,
            }
            response_raw_data = serialization.DictToBytes(json_response, encoding='utf-8')
            try:
                transport.write(response_raw_data)
            except:
                lg.exc()
                return False
            if _Debug:
                lg.out(_DebugLevel, '    wrote %d bytes to %s/%s at %r' % (len(response_raw_data), route_id, transport_direction, transport))
            return True
    call_id = json_data.get('call_id')
    if call_id or (cmd in (
        'api',
        'response',
        'push',
        'server-public-key',
        'client-public-key',
        'server-code',
        'client-code',
    )):
        route_info = routes(route_id)
        if _Debug:
            lg.args(_DebugLevel, direction=transport_direction, route_info=route_info)
        if not route_info:
            lg.warn('route info for %r was not found' % route_id)
            return False
        external_transport = route_info.get('external_transport')
        internal_transport = route_info.get('internal_transport')
        if transport_direction == 'external':
            if not internal_transport:
                if not external_transport:
                    lg.warn('internal and external transport for route %r are both disconnected' % route_id)
                    return False
                lg.warn('internal transport for route %r is not connected' % route_id)
                json_response = {
                    'cmd': 'server-disconnected',
                }
                response_raw_data = serialization.DictToBytes(json_response, encoding='utf-8')
                try:
                    transport.write(response_raw_data)
                except:
                    lg.exc()
                    return False
                return True
            try:
                internal_transport.write(raw_data)
            except:
                lg.exc()
                return False
            if _Debug:
                lg.out(_DebugLevel, '    wrote %d bytes to %s/%s at %r' % (len(raw_data), route_id, 'internal', internal_transport))
            return True
        if transport_direction == 'internal':
            if not external_transport:
                if not internal_transport:
                    lg.warn('external and internal transport for route %r are both disconnected' % route_id)
                    return False
                lg.warn('external transport for route %r is not connected' % route_id)
                json_response = {
                    'cmd': 'client-disconnected',
                }
                response_raw_data = serialization.DictToBytes(json_response, encoding='utf-8')
                try:
                    transport.write(response_raw_data)
                except:
                    lg.exc()
                    return False
                return True
            try:
                external_transport.write(raw_data)
            except:
                lg.exc()
                return False
            if _Debug:
                lg.out(_DebugLevel, '    wrote %d bytes to %s/%s at %r' % (len(raw_data), route_id, 'external', external_transport))
            return True
        lg.warn('unexpected direction: %r' % transport_direction)
    return False



"""
websocket - WebSocket client library for Python

Copyright (C) 2010 Hiroki Ohtani(liris)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA  02110-1335  USA

"""
"""
WebSocketApp provides higher level APIs.
"""
import inspect
import select
import sys
import threading
import time
import traceback

import six

from ._abnf import ABNF
from ._core import WebSocket, getdefaulttimeout
from ._exceptions import *
from . import _logging

__all__ = ['WebSocketApp']


class Dispatcher:
    def __init__(self, app, ping_timeout):
        self.app = app
        self.ping_timeout = ping_timeout

    def read(self, sock, read_callback, check_callback):
        while self.app.keep_running:
            r, w, e = select.select((self.app.sock.sock, ), (), (), self.ping_timeout)
            if r:
                if not read_callback():
                    break
            check_callback()


class SSLDispatcher:
    def __init__(self, app, ping_timeout):
        self.app = app
        self.ping_timeout = ping_timeout

    def read(self, sock, read_callback, check_callback):
        while self.app.keep_running:
            r = self.select()
            if r:
                if not read_callback():
                    break
            check_callback()

    def select(self):
        sock = self.app.sock.sock
        if sock.pending():
            return [
                sock,
            ]

        r, w, e = select.select((sock, ), (), (), self.ping_timeout)
        return r


class WebSocketApp(object):
    """
    Higher level of APIs are provided.
    The interface is like JavaScript WebSocket object.
    """
    def __init__(self, url, header=None, on_open=None, on_message=None, on_error=None, on_close=None, on_ping=None, on_pong=None, on_cont_message=None, keep_running=True, get_mask_key=None, cookie=None, subprotocols=None, on_data=None):
        """
        url: websocket url.
        header: custom header for websocket handshake.
        on_open: callable object which is called at opening websocket.
          this function has one argument. The argument is this class object.
        on_message: callable object which is called when received data.
         on_message has 2 arguments.
         The 1st argument is this class object.
         The 2nd argument is utf-8 string which we get from the server.
        on_error: callable object which is called when we get error.
         on_error has 2 arguments.
         The 1st argument is this class object.
         The 2nd argument is exception object.
        on_close: callable object which is called when closed the connection.
         this function has one argument. The argument is this class object.
        on_cont_message: callback object which is called when receive continued
         frame data.
         on_cont_message has 3 arguments.
         The 1st argument is this class object.
         The 2nd argument is utf-8 string which we get from the server.
         The 3rd argument is continue flag. if 0, the data continue
         to next frame data
        on_data: callback object which is called when a message received.
          This is called before on_message or on_cont_message,
          and then on_message or on_cont_message is called.
          on_data has 4 argument.
          The 1st argument is this class object.
          The 2nd argument is utf-8 string which we get from the server.
          The 3rd argument is data type. ABNF.OPCODE_TEXT or ABNF.OPCODE_BINARY will be came.
          The 4th argument is continue flag. if 0, the data continue
        keep_running: this parameter is obsolete and ignored.
        get_mask_key: a callable to produce new mask keys,
          see the WebSocket.set_mask_key's docstring for more information
        subprotocols: array of available sub protocols. default is None.
        """
        self.url = url
        self.header = header if header is not None else []
        self.cookie = cookie

        self.on_open = on_open
        self.on_message = on_message
        self.on_data = on_data
        self.on_error = on_error
        self.on_close = on_close
        self.on_ping = on_ping
        self.on_pong = on_pong
        self.on_cont_message = on_cont_message
        self.keep_running = False
        self.get_mask_key = get_mask_key
        self.sock = None
        self.last_ping_tm = 0
        self.last_pong_tm = 0
        self.subprotocols = subprotocols

    def __repr__(self):
        return 'WebSocketApp[%s|%s]' % (self.url, self.sock.connected if self.sock else '-')

    def send(self, data, opcode=ABNF.OPCODE_TEXT):
        """
        send message.
        data: message to send. If you set opcode to OPCODE_TEXT,
              data must be utf-8 string or unicode.
        opcode: operation code of data. default is OPCODE_TEXT.
        """
        if not self.sock or self.sock.send(data, opcode) == 0:
            raise WebSocketConnectionClosedException('Connection is already closed.')

    def close(self, **kwargs):
        """
        close websocket connection.
        """
        self.keep_running = False
        if self.sock:
            self.sock.close(**kwargs)
            self.sock = None

    def _send_ping(self, interval, event):
        while not event.wait(interval):
            self.last_ping_tm = time.time()
            if self.sock:
                try:
                    self.sock.ping()
                except Exception as ex:
                    _logging.warning('send_ping routine terminated: {}'.format(ex))
                    break

    def run_forever(
        self,
        sockopt=None,
        sslopt=None,
        ping_interval=0,
        ping_timeout=None,
        http_proxy_host=None,
        http_proxy_port=None,
        http_no_proxy=None,
        http_proxy_auth=None,
        skip_utf8_validation=False,
        host=None,
        origin=None,
        dispatcher=None,
        suppress_origin=False,
        proxy_type=None,
    ):
        """
        run event loop for WebSocket framework.
        This loop is infinite loop and is alive during websocket is available.
        sockopt: values for socket.setsockopt.
            sockopt must be tuple
            and each element is argument of sock.setsockopt.
        sslopt: ssl socket optional dict.
        ping_interval: automatically send "ping" command
            every specified period(second)
            if set to 0, not send automatically.
        ping_timeout: timeout(second) if the pong message is not received.
        http_proxy_host: http proxy host name.
        http_proxy_port: http proxy port. If not set, set to 80.
        http_no_proxy: host names, which doesn't use proxy.
        skip_utf8_validation: skip utf8 validation.
        host: update host header.
        origin: update origin header.
        dispatcher: customize reading data from socket.
        suppress_origin: suppress outputting origin header.

        Returns
        -------
        False if caught KeyboardInterrupt
        True if other exception was raised during a loop
        """

        if ping_timeout is not None and ping_timeout <= 0:
            ping_timeout = None
        if ping_timeout and ping_interval and ping_interval <= ping_timeout:
            raise WebSocketException('Ensure ping_interval > ping_timeout')
        if not sockopt:
            sockopt = []
        if not sslopt:
            sslopt = {}
        if self.sock:
            raise WebSocketException('socket is already opened')
        thread = None
        self.keep_running = True
        self.last_ping_tm = 0
        self.last_pong_tm = 0

        def teardown(close_frame=None):
            """
            Tears down the connection.
            If close_frame is set, we will invoke the on_close handler with the
            statusCode and reason from there.
            """
            if thread:
                isAlive = False
                if hasattr(thread, 'isAlive'):
                    isAlive = thread.isAlive()
                elif hasattr(thread, 'is_alive'):
                    isAlive = thread.is_alive()
                if isAlive:
                    event.set()
                    thread.join()
            self.keep_running = False
            if self.sock:
                self.sock.close()
            close_args = self._get_close_args(close_frame.data if close_frame else None)
            self._callback(self.on_close, *close_args)
            self.sock = None

        try:
            self.sock = WebSocket(self.get_mask_key, sockopt=sockopt, sslopt=sslopt, fire_cont_frame=self.on_cont_message is not None, skip_utf8_validation=skip_utf8_validation, enable_multithread=True if ping_interval else False)
            self.sock.settimeout(getdefaulttimeout())
            self.sock.connect(
                self.url,
                header=self.header,
                cookie=self.cookie,
                http_proxy_host=http_proxy_host,
                http_proxy_port=http_proxy_port,
                http_no_proxy=http_no_proxy,
                http_proxy_auth=http_proxy_auth,
                subprotocols=self.subprotocols,
                host=host,
                origin=origin,
                suppress_origin=suppress_origin,
                proxy_type=proxy_type,
            )
            if not dispatcher:
                dispatcher = self.create_dispatcher(ping_timeout)

            self._callback(self.on_open)

            if ping_interval:
                event = threading.Event()
                thread = threading.Thread(target=self._send_ping, args=(ping_interval, event))
                thread.setDaemon(True)
                thread.start()

            def read():
                if not self.keep_running:
                    return teardown()

                op_code, frame = self.sock.recv_data_frame(True)
                if op_code == ABNF.OPCODE_CLOSE:
                    return teardown(frame)
                elif op_code == ABNF.OPCODE_PING:
                    self._callback(self.on_ping, frame.data)
                elif op_code == ABNF.OPCODE_PONG:
                    self.last_pong_tm = time.time()
                    self._callback(self.on_pong, frame.data)
                elif op_code == ABNF.OPCODE_CONT and self.on_cont_message:
                    self._callback(self.on_data, frame.data, frame.opcode, frame.fin)
                    self._callback(self.on_cont_message, frame.data, frame.fin)
                else:
                    data = frame.data
                    if six.PY3 and op_code == ABNF.OPCODE_TEXT:
                        data = data.decode('utf-8')
                    self._callback(self.on_data, data, frame.opcode, True)
                    self._callback(self.on_message, data)

                return True

            def check():
                if (ping_timeout):
                    has_timeout_expired = time.time() - self.last_ping_tm > ping_timeout
                    has_pong_not_arrived_after_last_ping = self.last_pong_tm - self.last_ping_tm < 0
                    has_pong_arrived_too_late = self.last_pong_tm - self.last_ping_tm > ping_timeout

                    if (self.last_ping_tm and has_timeout_expired and (has_pong_not_arrived_after_last_ping or has_pong_arrived_too_late)):
                        raise WebSocketTimeoutException('ping/pong timed out')
                return True

            dispatcher.read(self.sock.sock, read, check)
        except (Exception, KeyboardInterrupt, SystemExit) as e:
            self._callback(self.on_error, e)
            if isinstance(e, SystemExit):
                # propagate SystemExit further
                raise
            teardown()
            return not isinstance(e, KeyboardInterrupt)

    def create_dispatcher(self, ping_timeout):
        timeout = ping_timeout or 10
        if self.sock.is_ssl():
            return SSLDispatcher(self, timeout)

        return Dispatcher(self, timeout)

    def _get_close_args(self, data):
        """ this functions extracts the code, reason from the close body
        if they exists, and if the self.on_close except three arguments """
        # if the on_close callback is "old", just return empty list
        if sys.version_info < (3, 0):
            if not self.on_close or len(inspect.getargspec(self.on_close).args) != 3:
                return []
        else:
            if not self.on_close or len(inspect.getfullargspec(self.on_close).args) != 3:
                return []

        if data and len(data) >= 2:
            code = 256*six.byte2int(data[0:1]) + six.byte2int(data[1:2])
            reason = data[2:].decode('utf-8')
            return [code, reason]

        return [None, None]

    def _callback(self, callback, *args):
        if callback:
            try:
                if inspect.ismethod(callback):
                    callback(*args)
                else:
                    callback(self, *args)

            except Exception as e:
                _logging.error('error from callback {}: {}'.format(callback, e))
                if _logging.isEnabledForDebug():
                    _, _, tb = sys.exc_info()
                    traceback.print_tb(tb)
